@@ -7,6 +7,7 @@ License: See project-level license file.
 
 import networkx as nx
 from . import function_composer
+import inspect
 
 
 starting_node_properties = {"type": "standard", "has_value": False, "value": None, "is_frozen": False, "is_recipe": False}
@@ -53,15 +54,13 @@ class Graph():
             raise ValueError("Cannot add node with dependencies without a recipe")
 
         elif recipe is None:  # Accept nodes with no dependencies
-            # Avoid adding existing node so as not to overwrite attributes
-            if name not in self.nodes:
-                self._nxdg.add_node(name, **starting_node_properties)
+            # The node is readded even if already existing
+            self._nxdg.add_node(name, **starting_node_properties)
 
         else:  # Standard case
             # Add the node
-            # Avoid adding existing node so as not to overwrite attributes
-            if name not in self.nodes:
-                self._nxdg.add_node(name, **starting_node_properties)
+            # The node is readded even if already existing
+            self._nxdg.add_node(name, **starting_node_properties)
             # Set attributes
             # Note: This could be done in the constructor, but doing it separately adds flexibility
             # Indeed, we might want to change how attributes work, and we can do it by modifying setters
@@ -117,6 +116,63 @@ class Graph():
         # Add possibilities to the list of possibilities of the conditional
         self.set_possibilities(name, [value_true, value_false])
 
+    def add_multiple_conditional(self, name, conditions, possibilities):
+        """
+        Interface to add a multiple conditional to the graph.
+        """
+        # Add all nodes and connect all edges
+        self._nxdg.add_node(name, **starting_node_properties)
+        for node in conditions + possibilities:
+            # Avoid adding existing dependencies so as not to overwrite attributes
+            if node not in self.nodes:
+                self._nxdg.add_node(node, **starting_node_properties)
+            self._nxdg.add_edge(node, name)
+
+        # Specify that this node is a conditional
+        self.set_type(name, "conditional")
+
+        # Add conditions name to the list of conditions of the conditional
+        self.set_conditions(name, conditions)
+
+        # Add possibilities to the list of possibilities of the conditional
+        self.set_possibilities(name, possibilities)
+
+    def edit_step(self, name, recipe=None, *args, **kwargs):
+        """
+        Interface to edit an existing node, changing its predecessors
+        """
+        if name not in self.nodes:
+            raise ValueError("Cannot edit non-existent node " + name)
+
+        # Store old attributes
+        was_recipe = self.is_recipe(name)
+        was_frozen = self.is_frozen(name)
+        had_value = self.has_value(name)
+        if had_value:
+            old_value = self.get_value(name)
+
+        # Remove in-edges from the node because we need to replace them
+        # use of list() is to make a copy because in_edges() returns a view
+        self._nxdg.remove_edges_from(list(self._nxdg.in_edges(name)))
+        # Readd the step. This should not break anything
+        self.add_step(name, recipe, *args, **kwargs)
+
+        # Readd attributes
+        # Readding out-edges is not needed because we never removed them
+        self.set_is_recipe(name, was_recipe)
+        self.set_is_frozen(name, was_frozen)
+        self.set_has_value(name, had_value)
+        if had_value:
+            self.set_value(name, old_value)
+
+    def remove_step(self, name):
+        """
+        Interface to remove an existing node, without changing anything else
+        """
+        if name not in self.nodes:
+            raise ValueError("Cannot edit non-existent node " + name)
+        self._nxdg.remove_node(name)
+
     def get_node_attribute(self, node, attribute):
         attributes = self.nodes[node]
         if attribute in attributes and attributes[attribute] is not None:
@@ -195,11 +251,16 @@ class Graph():
     def set_has_value(self, node, has_value):
         return self.set_node_attribute(node, "has_value", has_value)
 
-    def clear_values(self):
+    def clear_values(self, *args):
         """
-        Clear all values in the graph nodes.
+        Clear values in the graph nodes.
         """
-        for node in self.nodes:
+        if len(args) == 0:  # Interpret as "Clear everything"
+            nodes_to_clear = self.nodes
+        else:
+            nodes_to_clear = args & self.nodes  # Intersection
+
+        for node in nodes_to_clear:
             if self.is_frozen(node):
                 continue
             self.unset_value(node)
@@ -295,27 +356,28 @@ class Graph():
         """
         return {key: self.get_value(value) for key, value in dictionary.items()}
 
-    def evaluate_target(self, target):
+    def evaluate_target(self, target, continue_on_fail=False):
         """
         Generic interface to evaluate a GenericNode.
         """
         if self.get_type(target) == "standard":
-            return self.evaluate_standard(target)
+            return self.evaluate_standard(target, continue_on_fail)
         elif self.get_type(target) == "conditional":
-            return self.evaluate_conditional(target)
+            return self.evaluate_conditional(target, continue_on_fail)
         else:
             raise ValueError("Evaluation of nodes of type ", self.get_type(target), " is not supported")
 
-    def evaluate_standard(self, node):
+    def evaluate_standard(self, node, continue_on_fail=False):
         """
         Evaluate of a node.
         """
         # Check if it already has a value
         if self.has_value(node):
-            return self.get_value(node)
+            self.get_value(node)
+            return
         # If not, evaluate all arguments
         for dependency_name in self._nxdg.predecessors(node):
-            self.evaluate_target(dependency_name)
+            self.evaluate_target(dependency_name, continue_on_fail)
 
         # Actual computation happens here
         try:
@@ -323,40 +385,66 @@ class Graph():
             func = self.get_value(recipe)
             res = func(*self.get_list_of_values(self.get_args(node)), **self.get_kwargs_values(self.get_kwargs(node)))
         except Exception as e:
-            if len(e.args) > 0:
-                e.args = ("While evaluating " + node + ": " + str(e.args[0]),) + e.args[1:]
-            raise
+            if continue_on_fail:
+                # Do nothing, we want to keep going
+                return
+            else:
+                if len(e.args) > 0:
+                    e.args = ("While evaluating " + node + ": " + str(e.args[0]),) + e.args[1:]
+                raise
         # Save results
         self.set_value(node, res)
-        return res
 
-    def evaluate_conditional(self, conditional):
+    def evaluate_conditional(self, conditional, continue_on_fail=False):
         """
         Evaluate a conditional.
         """
         # Check if it already has a value
         if self.has_value(conditional):
-            return self.get_value(conditional)
+            self.get_value(conditional)
+            return
         # If not, evaluate the conditions until one is found true
         for index, condition in enumerate(self.get_conditions(conditional)):
-            res = self.evaluate_target(condition)
-            if res:
+            self.evaluate_target(condition, continue_on_fail)
+            if self.has_value(condition) and self.get_value(condition):
                 break
+            elif not self.has_value(condition):
+                # Computing failed
+                if continue_on_fail:
+                    # Do nothing, we want to keep going
+                    return
+                else:
+                    raise ValueError("Node " + condition + " could not be computed.")
         else:  # Happens if loop is never broken, i.e. when no conditions are true
             index = -1
 
         # Actual computation happens here
-        res = self.evaluate_target(self.get_possibilities(conditional)[index])
+        try:
+            possibility = self.get_possibilities(conditional)[index]
+            self.evaluate_target(possibility, continue_on_fail)
+            res = self.get_value(possibility)
+        except:
+            if continue_on_fail:
+                # Do nothing, we want to keep going
+                return
+            else:
+                raise ValueError("Node " + possibility + " could not be computed.")
         # Save results and release
         self.set_value(conditional, res)
-        return res
 
     def execute_to_targets(self, *targets):
         """
         Evaluate all nodes in the graph that are needed to reach the targets.
         """
         for target in targets:
-            self.evaluate_target(target)
+            self.evaluate_target(target, False)
+
+    def progress_towards_targets(self, *targets):
+        """
+        Move towards the targets by evaluating nodes, but keep going if evaluation fails.
+        """
+        for target in targets:
+            self.evaluate_target(target, True)
 
     def is_other_node_compatible(self, node, other, other_node):
         # If types differ, return False
@@ -367,9 +455,15 @@ class Graph():
             return True
         # If they both have values but they differ, return False. If only one has a value, proceed
         if self.has_value(node) and other.has_value(other_node) and self.get_value(node) != other.get_value(other_node):
-            return False
+            # Plot twist! Both are functions and have the same code: proceed
+            if inspect.isfunction(self.get_value(node)) and inspect.isfunction(other.get_value(other_node)) and self.get_value(node).__code__.co_code == other.get_value(other_node).__code__.co_code:
+                pass
+            else:
+                return False
         # If they both have dependencies but they differ, return False. If only one has dependencies, proceed
-        if len(list(self._nxdg.predecessors(node))) != 0 and len(list(other._nxdg.predecessors(other_node))) != 0 and self._nxdg.predecessors(node) != other._nxdg.predecessors(other_node):
+        predecessors = list(self._nxdg.predecessors(node))
+        other_predecessors = list(other._nxdg.predecessors(other_node))
+        if len(predecessors) != 0 and len(other_predecessors) != 0 and predecessors != other_predecessors:
             return False
         # Return True if at least one has no dependencies (or they are the same), at least one has no value (or they are the same)
         return True
@@ -406,6 +500,8 @@ class Graph():
         subfuncs_dependencies = []
         for argument in self.get_kwargs(node_name):
             if argument == dependency_name:
+                if self.get_type(dependency_name) == "conditional" or self.get_type(self.get_recipe(dependency_name)) == "conditional":
+                    raise TypeError("Simplification does not support conditional nodes.")
                 subfuncs.append(self[self.get_recipe(dependency_name)])  # Get python function
                 subfuncs_dependencies.append(list(self.get_args(dependency_name)) + list(self.get_kwargs(dependency_name).values()))
             else:
@@ -434,7 +530,7 @@ class Graph():
         if len(args) == 0:  # Interpret as "Freeze everything"
             nodes_to_freeze = self.nodes
         else:
-            nodes_to_freeze = args
+            nodes_to_freeze = args & self.nodes  # Intersection
 
         for key in nodes_to_freeze:
             if self.has_value(key):
@@ -444,7 +540,7 @@ class Graph():
         if len(args) == 0:  # Interpret as "Unfreeze everything"
             nodes_to_unfreeze = self.nodes.keys()
         else:
-            nodes_to_unfreeze = args
+            nodes_to_unfreeze = args & self.nodes  # Intersection
 
         for key in nodes_to_unfreeze:
             self.set_is_frozen(key, False)
@@ -453,7 +549,8 @@ class Graph():
         """
         Make dependencies (parents) of recipes also recipes
         """
-        for node in self.nodes:
+        # Work in reverse topological order, to get children before parents
+        for node in reversed(self.get_topological_order()):
             if self.is_recipe(node):
                 for parent in self._nxdg.predecessors(node):
                     self.set_is_recipe(parent, True)
@@ -467,3 +564,9 @@ class Graph():
         """
         self.make_recipe_dependencies_also_recipes()
         self.freeze()
+
+    def get_topological_order(self):
+        """
+        Return list of nodes in topological order, i.e., from dependencies to targets
+        """
+        return list(nx.topological_sort(self._nxdg))
